# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
package Storage::ArchiveStorage;

use strict;
use warnings;

use Fcntl qw(F_GETFL F_SETFL O_NONBLOCK);
use File::Basename qw(basename);
use IO::Select;
use IPC::Open3;
use IPC::Run;
use List::Util qw(sum);
use HelpFuncs;
use Logging;
use Storage::Storage;

use vars qw|@ISA|;
@ISA = qw|Storage::Storage|;

sub _init {
  my ($self, %options) = @_;

  $self->SUPER::_init(%options);
  $self->{exportDir} = $options{'export-dir'};
  $self->{ftp} = $options{'ftp'};
  $self->{sessionPath} = $options{'session-path'};
  $self->{verbose} = $options{'verbose'};
  $self->{keepLocalBackup} = $options{'keep-local-backup'};
  $self->{keepLocalBackupIfExportFailed} = $options{'keep-local-backup-if-export-failed'};

  $self->{pmmRasPid} = undef;
  $self->{pmmRasHandles} = [undef, undef, undef];
  $self->{pmmRasInputSelect} = undef;
  $self->{pmmRasOutputSelect} = undef;
  $self->{pmmRasInput} = '';
  $self->{pmmRasOutput} = '';

  $self->{currentSize} = [];

  Logging::debug("-" x 60);
  Logging::debug("ARCHIVE storage initialized.");
  Logging::debug("Base directory: $self->{output_dir}");
  Logging::debug("Space reserved: $self->{space_reserved}");
  Logging::debug("Gzip bundles: " . ($self->{gzip_bundle} ? "yes" : "no"));
  Logging::debug("Bundle split size: " . ($self->{split_size} || "do not split"));
  Logging::debug("-" x 60);
  $self->reserveSpace();
}

sub finishExport {
  my $self = shift;

  my $mainDumpFile = $self->getMainDumpXmlRelativePath();
  if ($mainDumpFile) {
    eval {
      $self->_addFile(basename($mainDumpFile));
    };
    if ($@) {
      Logging::debug($@);
    }
  }
  eval {
    $self->_finishPmmRas();
    if (!$self->{keepLocalBackup}) {
      $self->_deleteLocalBackup();
    }
  };
  if ($@) {
    if ($self->{keepLocalBackup} || $self->{keepLocalBackupIfExportFailed}) {
      Logging::warning($@, undef, 'step' => 'upload-to-ftp');
      $self->_deleteDumpFiles();
    } else {
      die($@);
    }
  }
}

sub writeDiscovered{
  my ( $self, $dumpPath, $dumpXmlName, $dumpSize, $ownerGuid, $ownerType, $objectGuid, $objectId ) = @_;

  $self->_writeDiscovered($dumpPath, $dumpXmlName, $dumpSize, $ownerGuid, $ownerType, $objectGuid, $objectId);

  Logging::debug("Content $self->{exportDir}");
  my ($content) = $self->{exportDir} =~ m/^ext:\/\/([^\/]+)\/.*/;
  return unless defined $content;

  $content .= "\nlocal";
  Logging::debug("Content $content");

  my $idx = rindex( $dumpXmlName, '.xml' );
  $dumpXmlName = substr( $dumpXmlName, 0, $idx ) if $idx>0;
  my $destDir = $self->getFullOutputPath();
  $destDir .= "/$dumpPath" if $dumpPath;
  $destDir .= "/.discovered/$dumpXmlName";
  Logging::debug("Writing $destDir/storages");
  open(my $handle, '>', "$destDir/storages");
  print $handle $content;
  close($handle);

  my @files = (['storages', length($content)]);
  $self->regIdFiles( $destDir, $destDir, length($content), \@files );
  return [$destDir, \@files];
}

sub moveFileToDiscovered {
  my $self = shift;

  $self->_moveFileToDiscovered(@_);
}

sub addTar {
  my $self = shift;

  my ($ret) = @{$self->_addTar(@_)};
  return $ret;
}

sub onVolumeCreate() {
  my ($self, $destDir, $fileName) = @_;

  my $file;
  eval {
    my $dumpStorage = $self->{mainDumpRootPath} ? sprintf('%s/%s', $self->{output_dir}, $self->{mainDumpRootPath}) : $self->{output_dir};
    my $destDirRelativePath = substr($destDir, length($dumpStorage));
    $destDirRelativePath =~ s/^\/+//;
    $file = $destDirRelativePath ? sprintf('%s/%s', $destDirRelativePath, $fileName) : $fileName;
    $self->_addFile($file);
  };
  if ($@) {
    Logging::debug($@);
    unless ($self->{keepLocalBackup} || $self->{keepLocalBackupIfExportFailed}) {
      my $pmmRasOutput = HelpFuncs::trim($self->{pmmRasOutput});
      Logging::error(sprintf('Unable to upload the archive %s to the external storage: %s', $file, $pmmRasOutput), undef, 'step' => 'upload-to-ftp');
      if ($self->{sessionPath}) {
        Logging::serializeXmlLog("$self->{sessionPath}/migration.result");
      }
      exit(0);
    }
  }
}

sub CleanupFiles()
{
  my $self = shift;

  eval {
    $self->_finishPmmRas();
  };
  if ($@) {
    Logging::debug($@);
  }
  $self->SUPER::CleanupFiles();
  $self->_deleteDumpFiles();
}

sub _deleteLocalBackup {
  my $self = shift;

  Logging::debug("Delete a local backup");
  eval {
    my $command = [AgentConfig::pmmRasBin(), '--delete-dump'];
    push(@{$command}, '--dump-specification', $self->getMainDumpXmlRelativePath());
    push(@{$command}, '--session-path', $self->{sessionPath}) if $self->{sessionPath};
    push(@{$command}, '--debug', '--verbose') if $self->{verbose};

    Logging::debug('Exec: '.join(' ', @{$command}));
    my $stdout;
    IPC::Run::run($command, '1>', \$stdout) or die($stdout);
  };
  if ($@) {
    Logging::debug('Unable to delete a local backup: ' . HelpFuncs::trim($@));
  }
}

sub _deleteDumpFiles {
  my $self = shift;

  Logging::debug("Delete dump files");
  eval {
    my $command = [AgentConfig::pmmRasBin(), '--delete-dump-files'];
    if ($self->{ftp}) {
      $ENV{'DUMP_STORAGE_PASSWD'} = $self->{ftp}->{password};
      push(@{$command}, '--dump-storage', HelpFuncs::getStorageUrlFromFtpSettings($self->{ftp}));
      push(@{$command}, '--dump-file-specification', $self->{exportFileName});
      push(@{$command}, '--use-ftp-passive-mode') if exists($self->{ftp}->{'passive'});
    } else {
      push(@{$command}, '--dump-storage', $self->{exportDir});
      push(@{$command}, '--dump-file-specification', $self->{exportFileName});
    }
    push(@{$command}, '--session-path', $self->{sessionPath}) if $self->{sessionPath};
    push(@{$command}, '--debug', '--verbose') if $self->{verbose};

    Logging::debug('Exec: '.join(' ', @{$command}));
    my $stdout;
    IPC::Run::run($command, '1>', \$stdout) or die($stdout);
  };
  if ($@) {
    Logging::debug('Unable to delete dump files: ' . HelpFuncs::trim($@));
  }
}
sub _addFile {
  my ($self, $filePath) = @_;

  Logging::debug("Add file $filePath to archive");

  $self->_startPmmRas() unless $self->{pmmRasPid};

  my $sizeLimit = $self->{split_size} ? $self->{split_size} : $self->{maxArchiveSize};
  my $dumpStoragePath = $self->{mainDumpRootPath} ? "$self->{output_dir}/$self->{mainDumpRootPath}" : $self->{output_dir};
  my $fileSize = -s ("$dumpStoragePath/$filePath");
  push @{$self->{currentSize}}, $fileSize;
  my $currentSize = sum(0, @{$self->{currentSize}});
  Logging::debug("Used local space: $currentSize: Limit: $sizeLimit: +$fileSize");

  $self->{pmmRasInput} .= "$filePath\n";

  my $timeout = $self->{keepLocalBackup} || $self->{keepLocalBackupIfExportFailed} || $currentSize <= $sizeLimit ? 0 : undef; # set (un)blocking mode
  while (1) {
    my @result = IO::Select->select(
        $self->{pmmRasOutputSelect},
        0 < length($self->{pmmRasInput}) ? $self->{pmmRasInputSelect} : undef,
        undef,
        $timeout);
    my ($outputHandles, $inputHandles) = @result ? @result : ([], []);

    last if (0 == length($self->{pmmRasInput}) || 0 == scalar(@{$inputHandles})) && 0 == scalar(@{$outputHandles}) && defined($timeout);

    my $inputError = undef;
    if (0 < length($self->{pmmRasInput}) && 0 < scalar(@{$inputHandles})) {
      my $written = syswrite($self->{pmmRasHandles}->[0], $self->{pmmRasInput});
      if (defined($written)) {
        $self->{pmmRasInput} = length($self->{pmmRasInput}) == $written ? '' : substr($self->{pmmRasInput}, $written);
      } else {
        $inputError = "Failed to write pmm-ras input: $!";
      }
    }

    if (0 < scalar(@{$outputHandles})) {
      my $output;
      my $read = sysread($self->{pmmRasHandles}->[1], $output, 4096);
      die("Failed to read pmm-ras output: $!") unless defined($read);
      die("pmm-ras has stopped unexpectedly") if $read == 0;
      $self->{pmmRasOutput} .= $output;
      foreach my $char (split(//, $output)) {
        last unless "\n" eq $char;
        $fileSize = shift @{$self->{currentSize}};
        $currentSize = sum(0, @{$self->{currentSize}});
        Logging::debug("Used local space: $currentSize: Limit: $sizeLimit: -$fileSize");
      }
      $timeout = $self->{keepLocalBackup} || $self->{keepLocalBackupIfExportFailed} || $currentSize <= $sizeLimit ? 0 : undef; # reset (un)blocking mode
    }

    die($inputError) if defined($inputError);
  }
}

sub _startPmmRas {
  my $self = shift;

  my @command = (AgentConfig::pmmRasBin(), '--export-dump-as-file');
  if ($self->{ftp}) {
    $ENV{'DUMP_STORAGE_PASSWD'} = $self->{ftp}->{password};
    push(@command, '--dump-file-specification', sprintf('%s/%s', HelpFuncs::getStorageUrlFromFtpSettings($self->{ftp}), $self->{exportFileName}));
    push(@command, '--use-ftp-passive-mode') if exists($self->{ftp}->{'passive'});
  } else {
    push(@command, '--dump-file-specification', sprintf('%s/%s', $self->{exportDir}, $self->{exportFileName}));
  }
  push(@command, '--dump-storage', sprintf('%s/%s', $self->{output_dir}, $self->{mainDumpRootPath}));
  push(@command, '--remove-dump') unless $self->{keepLocalBackup} || $self->{keepLocalBackupIfExportFailed};
  push(@command, '--split-size', $self->{split_size}) if $self->{split_size};
  push(@command, '--session-path', $self->{sessionPath}) if $self->{sessionPath};
  push(@command, '--debug', '--verbose') if $self->{verbose};

  open($self->{pmmRasHandles}->[2], '>', '/dev/null') || die "Failed to open /dev/null: $!";

  Logging::debug("Start: @command");
  $self->{pmmRasPid} = open3(
      $self->{pmmRasHandles}->[0],
      $self->{pmmRasHandles}->[1],
      '>&' . fileno($self->{pmmRasHandles}->[2]),
      @command
  );

  _setBlockingMode($self->{pmmRasHandles}->[0], 0);
  _setBlockingMode($self->{pmmRasHandles}->[1], 0);

  $self->{pmmRasInputSelect} = IO::Select->new($self->{pmmRasHandles}->[0]);
  $self->{pmmRasOutputSelect} = IO::Select->new($self->{pmmRasHandles}->[1]);
}

sub _finishPmmRas {
  my $self = shift;

  return unless $self->{pmmRasPid};

  _setBlockingMode($self->{pmmRasHandles}->[0], 1);
  _setBlockingMode($self->{pmmRasHandles}->[1], 1);

  close($self->{pmmRasHandles}->[0]);

  my $output;
  while (sysread($self->{pmmRasHandles}->[1], $output, 4096)) {
    $self->{pmmRasOutput} .= $output;
  }
  my $pmmRasOutput = HelpFuncs::trim($self->{pmmRasOutput});

  waitpid($self->{pmmRasPid}, 0);
  my $exitCode = $? >> 8;

  close($self->{pmmRasHandles}->[1]);
  close($self->{pmmRasHandles}->[2]);

  $self->{pmmRasPid} = undef;
  $self->{pmmRasHandles} = [undef, undef, undef];
  $self->{pmmRasInputSelect} = undef;
  $self->{pmmRasOutputSelect} = undef;
  $self->{pmmRasOutput} = '';

  if ($exitCode) {
    die(sprintf('Unable to upload the backup to the external storage: %s', $pmmRasOutput));
  }

  my @createdVolumes = $pmmRasOutput ? split("\n", $pmmRasOutput) : ();
  $self->{createdVolumes} = \@createdVolumes;
}

sub _setBlockingMode {
  my ($handler, $blocking) = @_;

  my $flags = fcntl($handler, F_GETFL, 0) or die "Failed to get flags for the handler: $!";
  $flags = $blocking ? $flags & ~O_NONBLOCK : $flags | O_NONBLOCK;
  fcntl($handler, F_SETFL, $flags) or die "Failed to set flags for the handler: $!";
}

1;
